from snowddl.blueprint import SchemaBlueprint, DatabaseBlueprint
from snowddl.resolver.abc_resolver import AbstractResolver, ResolveResult, ObjectType, SnowDDLUnsupportedError


class SchemaResolver(AbstractResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.SCHEMA

    def get_existing_objects(self):
        return self.engine.schema_cache.schemas

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(SchemaBlueprint)

    def create_object(self, bp: SchemaBlueprint):
        query = self.engine.query_builder()
        query.append("CREATE")

        if bp.is_transient:
            query.append("TRANSIENT")

        query.append("SCHEMA {full_name:i}", {"full_name": bp.full_name})

        if self.config.get_permission_model(bp.permission_model).ruleset.create_managed_access_schema:
            query.append_nl("WITH MANAGED ACCESS")

        if bp.retention_time is not None:
            query.append_nl("DATA_RETENTION_TIME_IN_DAYS = {retention_time:d}", {"retention_time": bp.retention_time})

        if bp.comment:
            query.append_nl(
                "COMMENT = {comment}",
                {
                    "comment": bp.comment,
                },
            )

        self.engine.execute_safe_ddl(query)

        return ResolveResult.CREATE

    def compare_object(self, bp: SchemaBlueprint, row: dict):
        if bp.is_transient != row["is_transient"]:
            if bp.is_transient:
                raise SnowDDLUnsupportedError(f"Cannot change PERMANENT schema [{bp.full_name}] into TRANSIENT schema")
            else:
                raise SnowDDLUnsupportedError(f"Cannot change TRANSIENT schema [{bp.full_name}] into PERMANENT schema")

        query = self.engine.query_builder()

        query.append(
            "ALTER SCHEMA {full_name:i} SET ",
            {
                "full_name": bp.full_name,
            },
        )

        result = ResolveResult.NOCHANGE

        if bp.retention_time is not None and bp.retention_time != row["retention_time"]:
            self.engine.execute_unsafe_ddl(
                "ALTER SCHEMA {full_name:i} SET DATA_RETENTION_TIME_IN_DAYS = {retention_time:d}",
                {
                    "full_name": bp.full_name,
                    "retention_time": bp.retention_time,
                },
            )

            result = ResolveResult.ALTER

        if bp.comment != row["comment"]:
            self.engine.execute_safe_ddl(
                "ALTER SCHEMA {full_name:i} SET COMMENT = {comment}",
                {
                    "full_name": bp.full_name,
                    "comment": bp.comment,
                },
            )

            result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl(
            "DROP SCHEMA {database:i}.{schema:i}",
            {
                "database": row["database"],
                "schema": row["schema"],
            },
        )

        return ResolveResult.DROP

    def _post_process(self):
        for result in self.resolved_objects.values():
            if result != ResolveResult.NOCHANGE:
                # Reload cache if at least one object was changed
                self.engine.schema_cache.reload()
                break

    def _resolve_drop(self):
        tasks = {}

        for schema_full_name in sorted(self.existing_objects):
            # Object exists in blueprints, should not be dropped
            if schema_full_name in self.blueprints:
                continue

            # Parent object is going to be dropped
            if self.engine.intention_cache.check_parent_drop_intention(self.object_type, schema_full_name):
                continue

            database_full_name = ".".join(schema_full_name.split(".")[:1])
            database_bp = self.config.get_blueprints_by_type(DatabaseBlueprint).get(database_full_name)

            # Schema database does not exist in blueprints, schema will be dropped automatically on DROP DATABASE
            if database_bp is None:
                continue

            # Schemas without blueprints are allowed in sandbox databases, should not be dropped
            if database_bp.is_sandbox:
                continue

            tasks[schema_full_name] = (self.drop_object, self.existing_objects[schema_full_name])

        self._process_tasks(tasks)



from snowddl.blueprint import DatabaseBlueprint, DatabaseIdent
from snowddl.parser.abc_parser import AbstractParser


# fmt: off
database_json_schema = {
    "type": "object",
    "properties": {
        "permission_model": {
            "type": "string",
        },
        "is_transient": {
            "type": "boolean"
        },
        "retention_time": {
            "type": "integer"
        },
        "is_sandbox": {
            "type": "boolean"
        },
        "owner_share_read": {
            "share_read": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
        },
        "owner_integration_usage": {
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "owner_warehouse_usage": {
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "owner_account_grants": {
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "owner_global_roles": {
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "comment": {
            "type": "string"
        }
    },
    "additionalProperties": False
}
# fmt: on


class DatabaseParser(AbstractParser):
    def load_blueprints(self):
        for database_path in self.base_path.iterdir():
            if not database_path.is_dir():
                continue

            # Skip special sub-directories
            if database_path.name.startswith("__"):
                continue

            database_name = database_path.name.upper()
            database_params = self.parse_single_file(database_path / "params.yaml", database_json_schema)

            # fmt: off
            databases_permission_model_name = database_params.get("permission_model", self.config.DEFAULT_PERMISSION_MODEL).upper()
            database_permission_model = self.config.get_permission_model(databases_permission_model_name)
            # fmt: on

            if not database_permission_model.ruleset.create_database_owner_role:
                for k in database_params:
                    if k.startswith("owner_"):
                        raise ValueError(
                            f"Cannot use parameter [{k}] for database [{database_name}], it should be configured on schema level"
                        )

            owner_additional_grants = []
            owner_additional_account_grants = []

            for share_name in database_params.get("owner_share_read", []):
                owner_additional_grants.append(self.build_share_role_grant(share_name))
                self.config.add_blueprint(self.build_share_role_blueprint(share_name))

            for integration_name in database_params.get("owner_integration_usage", []):
                owner_additional_grants.append(self.build_integration_usage_grant(integration_name))

            for warehouse_name in database_params.get("owner_warehouse_usage", []):
                owner_additional_grants.append(self.build_warehouse_role_grant(warehouse_name, self.config.USAGE_ROLE_TYPE))

            for account_grant in database_params.get("owner_account_grants", []):
                owner_additional_account_grants.append(self.build_account_grant(account_grant))

            for global_role_name in database_params.get("owner_global_roles", []):
                owner_additional_grants.append(self.build_global_role_grant(global_role_name))

            bp = DatabaseBlueprint(
                full_name=DatabaseIdent(self.env_prefix, database_name),
                permission_model=databases_permission_model_name,
                is_transient=database_params.get("is_transient", False),
                retention_time=database_params.get("retention_time", None),
                is_sandbox=database_params.get("is_sandbox", False),
                owner_additional_grants=owner_additional_grants,
                owner_additional_account_grants=owner_additional_account_grants,
                comment=database_params.get("comment", None),
            )

            self.config.add_blueprint(bp)



from enum import Enum


class ObjectType(Enum):
    # Technical object type, used for GRANTs only
    # There is no blueprint
    ACCOUNT = {
        "singular": "ACCOUNT",
        "plural": "ACCOUNTS",
    }

    ACCOUNT_PARAMETER = {
        "singular": "ACCOUNT PARAMETER",
        "plural": "ACCOUNT PARAMETERS",
        "blueprint_cls": "AccountParameterBlueprint",
    }

    AGGREGATION_POLICY = {
        "singular": "AGGREGATION POLICY",
        "plural": "AGGREGATION POLICIES",
        "blueprint_cls": "AggregationPolicyBlueprint",
    }

    ALERT = {
        "singular": "ALERT",
        "plural": "ALERTS",
        "is_future_grant_supported": True,
        "blueprint_cls": "AlertBlueprint",
    }

    CLONE_TABLE = {
        "singular": "CLONE TABLE",
        "plural": "CLONE TABLES",
        "blueprint_cls": "TableBlueprint",
    }

    DATABASE = {
        "singular": "DATABASE",
        "plural": "DATABASES",
        "blueprint_cls": "DatabaseBlueprint",
    }

    # Technical object type, used for GRANTs only
    # There is no blueprint
    DATABASE_ROLE = {
        "singular": "DATABASE ROLE",
        "plural": "DATABASE ROLES",
    }

    DYNAMIC_TABLE = {
        "singular": "DYNAMIC TABLE",
        "plural": "DYNAMIC TABLES",
        "singular_for_ref": "TABLE",
        "is_future_grant_supported": True,
        "blueprint_cls": "DynamicTableBlueprint",
    }

    EVENT_TABLE = {
        "singular": "EVENT TABLE",
        "plural": "EVENT TABLES",
        "singular_for_ref": "TABLE",
        "is_future_grant_supported": True,
        "blueprint_cls": "EventTableBlueprint",
    }

    EXTERNAL_ACCESS_INTEGRATION = {
        "singular": "EXTERNAL ACCESS INTEGRATION",
        "plural": "EXTERNAL ACCESS INTEGRATIONS",
        "singular_for_ref": "INTEGRATION",
        "singular_for_grant": "INTEGRATION",
        "blueprint_cls": "ExternalAccessIntegrationBlueprint",
    }

    EXTERNAL_FUNCTION = {
        "singular": "EXTERNAL FUNCTION",
        "plural": "EXTERNAL FUNCTIONS",
        "is_future_grant_supported": True,
        "blueprint_cls": "ExternalFunctionBlueprint",
    }

    EXTERNAL_TABLE = {
        "singular": "EXTERNAL TABLE",
        "plural": "EXTERNAL TABLES",
        "singular_for_ref": "TABLE",
        "is_future_grant_supported": True,
        "blueprint_cls": "ExternalTableBlueprint",
    }

    FILE_FORMAT = {
        "singular": "FILE FORMAT",
        "plural": "FILE FORMATS",
        "is_future_grant_supported": True,
        "blueprint_cls": "FileFormatBlueprint",
    }

    FUNCTION = {
        "singular": "FUNCTION",
        "plural": "FUNCTIONS",
        "is_future_grant_supported": True,
        "is_overloading_supported": True,
        "blueprint_cls": "FunctionBlueprint",
    }

    HYBRID_TABLE = {
        "singular": "HYBRID TABLE",
        "plural": "HYBRID TABLES",
        "singular_for_ref": "TABLE",
        "is_future_grant_supported": True,
        "blueprint_cls": "HybridTableBlueprint",
    }

    # Technical object type, used for GRANTs only
    # There is no blueprint
    INTEGRATION = {
        "singular": "INTEGRATION",
        "plural": "INTEGRATIONS",
    }

    MASKING_POLICY = {
        "singular": "MASKING POLICY",
        "plural": "MASKING POLICIES",
        "blueprint_cls": "MaskingPolicyBlueprint",
    }

    MATERIALIZED_VIEW = {
        "singular": "MATERIALIZED VIEW",
        "plural": "MATERIALIZED VIEWS",
        "singular_for_ref": "VIEW",
        "is_future_grant_supported": True,
        "blueprint_cls": "MaterializedViewBlueprint",
    }

    NETWORK_POLICY = {
        "singular": "NETWORK POLICY",
        "plural": "NETWORK POLICIES",
        "blueprint_cls": "NetworkPolicyBlueprint",
    }

    NETWORK_RULE = {
        "singular": "NETWORK RULE",
        "plural": "NETWORK RULES",
        "is_future_grant_supported": True,
        "blueprint_cls": "NetworkRuleBlueprint",
    }

    NOTEBOOK = {
        "singular": "NOTEBOOK",
        "plural": "NOTEBOOKS",
    }

    PIPE = {
        "singular": "PIPE",
        "plural": "PIPES",
        "is_future_grant_supported": True,
        "blueprint_cls": "PipeBlueprint",
    }

    PROCEDURE = {
        "singular": "PROCEDURE",
        "plural": "PROCEDURES",
        "is_future_grant_supported": True,
        "is_overloading_supported": True,
        "blueprint_cls": "ProcedureBlueprint",
    }

    PROJECTION_POLICY = {
        "singular": "PROJECTION POLICY",
        "plural": "PROJECTION POLICIES",
        "blueprint_cls": "ProjectionPolicyBlueprint",
    }

    RESOURCE_MONITOR = {
        "singular": "RESOURCE MONITOR",
        "plural": "RESOURCE MONITORS",
        "blueprint_cls": "ResourceMonitorBlueprint",
    }

    ROLE = {
        "singular": "ROLE",
        "plural": "ROLES",
        "blueprint_cls": "RoleBlueprint",
    }

    ROW_ACCESS_POLICY = {
        "singular": "ROW ACCESS POLICY",
        "plural": "ROW ACCESS POLICIES",
        "blueprint_cls": "RowAccessPolicyBlueprint",
    }

    SCHEMA = {
        "singular": "SCHEMA",
        "plural": "SCHEMAS",
        "is_future_grant_supported": True,
        "blueprint_cls": "SchemaBlueprint",
    }

    SECRET = {
        "singular": "SECRET",
        "plural": "SECRETS",
        "is_future_grant_supported": True,
        "blueprint_cls": "SecretBlueprint",
    }

    SEQUENCE = {
        "singular": "SEQUENCE",
        "plural": "SEQUENCES",
        "is_future_grant_supported": True,
        "blueprint_cls": "SequenceBlueprint",
    }

    SHARE = {
        "singluar": "SHARE",
        "plural": "SHARES",
        "blueprint_cls": "OutboundShareBlueprint",
    }

    STAGE = {
        "singular": "STAGE",
        "plural": "STAGES",
        "is_future_grant_supported": True,
        "blueprint_cls": "StageBlueprint",
    }

    STAGE_FILE = {
        "singular": "STAGE FILE",
        "plural": "STAGE FILES",
        "blueprint_cls": "StageFileBlueprint",
    }

    STREAM = {
        "singular": "STREAM",
        "plural": "STREAMS",
        "is_future_grant_supported": True,
        "blueprint_cls": "StreamBlueprint",
    }

    TABLE = {
        "singular": "TABLE",
        "plural": "TABLES",
        "is_future_grant_supported": True,
        "blueprint_cls": "TableBlueprint",
    }

    TAG = {
        "singular": "TAG",
        "plural": "TAGS",
        "is_future_grant_supported": True,
        "blueprint_cls": "TagBlueprint",
    }

    TASK = {
        "singular": "TASK",
        "plural": "TASKS",
        "is_future_grant_supported": True,
        "blueprint_cls": "TaskBlueprint",
    }

    VIEW = {
        "singular": "VIEW",
        "plural": "VIEWS",
        "is_future_grant_supported": True,
        "blueprint_cls": "ViewBlueprint",
    }

    USER = {
        "singular": "USER",
        "plural": "USERS",
        "blueprint_cls": "UserBlueprint",
    }

    WAREHOUSE = {
        "singular": "WAREHOUSE",
        "plural": "WAREHOUSES",
        "blueprint_cls": "WarehouseBlueprint",
    }

    # Constraints
    PRIMARY_KEY = {
        "singular": "PRIMARY KEY",
        "plural": "PRIMARY KEYS",
        "blueprint_cls": "PrimaryKeyBlueprint",
    }

    UNIQUE_KEY = {
        "singular": "UNIQUE KEY",
        "plural": "UNIQUE KEYS",
        "blueprint_cls": "UniqueKeyBlueprint",
    }

    FOREIGN_KEY = {
        "singular": "FOREIGN KEY",
        "plural": "FOREIGN KEYS",
        "blueprint_cls": "ForeignKeyBlueprint",
    }

    @property
    def singular(self):
        return self.value.get("singular")

    @property
    def plural(self):
        return self.value.get("plural")

    @property
    def singular_for_grant(self):
        return self.value.get("singular_for_grant", self.value.get("singular"))

    @property
    def singular_for_ref(self):
        return self.value.get("singular_for_ref", self.value.get("singular"))

    @property
    def blueprint_cls(self):
        # This import prevents cicrular dependency between blueprints and object types
        from . import blueprint

        return getattr(blueprint, self.value.get("blueprint_cls"))

    @property
    def is_future_grant_supported(self) -> bool:
        return self.value.get("is_future_grant_supported", False)

    @property
    def is_overloading_supported(self) -> bool:
        return self.value.get("is_overloading_supported", False)

    def __repr__(self):
        return f"<{self.__class__.__name__}.{super().name}>"



from .blueprint import (
    AbstractBlueprint,
    SchemaObjectBlueprint,
    DependsOnMixin,
    RoleBlueprint,
    AccountParameterBlueprint,
    AggregationPolicyBlueprint,
    AlertBlueprint,
    BusinessRoleBlueprint,
    DatabaseBlueprint,
    DatabaseRoleBlueprint,
    DynamicTableBlueprint,
    EventTableBlueprint,
    ExternalAccessIntegrationBlueprint,
    ExternalFunctionBlueprint,
    ExternalTableBlueprint,
    FileFormatBlueprint,
    ForeignKeyBlueprint,
    FunctionBlueprint,
    HybridTableBlueprint,
    MaterializedViewBlueprint,
    MaskingPolicyBlueprint,
    NetworkPolicyBlueprint,
    NetworkRuleBlueprint,
    OutboundShareBlueprint,
    PipeBlueprint,
    PrimaryKeyBlueprint,
    ProcedureBlueprint,
    ProjectionPolicyBlueprint,
    ResourceMonitorBlueprint,
    RowAccessPolicyBlueprint,
    SchemaBlueprint,
    SchemaRoleBlueprint,
    SecretBlueprint,
    SequenceBlueprint,
    ShareRoleBlueprint,
    StageBlueprint,
    StageFileBlueprint,
    StreamBlueprint,
    TableBlueprint,
    TagBlueprint,
    TaskBlueprint,
    TechnicalRoleBlueprint,
    UniqueKeyBlueprint,
    UserBlueprint,
    ViewBlueprint,
    WarehouseBlueprint,
    T_Blueprint,
)

from .column import (
    DynamicTableColumn,
    ExternalTableColumn,
    TableColumn,
    ViewColumn,
    ArgumentWithType,
    NameWithType,
    SearchOptimizationItem,
)
from .data_type import BaseDataType, DataType
from .edition import Edition
from .grant import Grant, AccountGrant, FutureGrant

from .ident import (
    AbstractIdent,
    AbstractIdentWithPrefix,
    Ident,
    AccountIdent,
    AccountObjectIdent,
    DatabaseIdent,
    DatabaseRoleIdent,
    OutboundShareIdent,
    SchemaIdent,
    SchemaObjectIdent,
    SchemaObjectIdentWithArgs,
    StageFileIdent,
    TableConstraintIdent,
)

from .ident_builder import (
    build_schema_object_ident,
    build_role_ident,
    build_grant_name_ident,
    build_future_grant_name_ident,
    build_default_namespace_ident,
)

from .object_type import ObjectType
from .permission_model import PermissionModel, PermissionModelCreateGrant, PermissionModelFutureGrant, PermissionModelRuleset
from .reference import (
    AggregationPolicyReference,
    ForeignKeyReference,
    IndexReference,
    MaskingPolicyReference,
    ProjectionPolicyReference,
    RowAccessPolicyReference,
    TagReference,
)
from .stage import StageWithPath, StageUploadFile



from snowddl.blueprint import ViewBlueprint

from snowddl.error import SnowDDLExecuteError
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType


class ViewResolver(AbstractSchemaObjectResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.VIEW

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW VIEWS IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            if r["is_materialized"] == "true":
                continue

            existing_objects[f"{r['database_name']}.{r['schema_name']}.{r['name']}"] = {
                "database": r["database_name"],
                "schema": r["schema_name"],
                "name": r["name"],
                "owner": r["owner"],
                "text": r["text"],
                "is_secure": r["is_secure"] == "true",
                "comment": r["comment"] if r["comment"] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(ViewBlueprint)

    def create_object(self, bp: ViewBlueprint):
        self.engine.execute_safe_ddl(self._build_create_view(bp))

        # Comments on views are broken and must be applied separately
        if bp.comment:
            self.engine.execute_safe_ddl(
                "COMMENT ON VIEW {full_name:i} IS {comment}",
                {
                    "full_name": bp.full_name,
                    "comment": bp.comment,
                },
            )

        return ResolveResult.CREATE

    def compare_object(self, bp: ViewBlueprint, row: dict):
        query = self._build_create_view(bp)

        # If view text is exactly the same
        if row["text"] == str(query):
            try:
                # ... and it is possible to query view (underlying objects were not changed)
                self.engine.describe_meta(
                    "SELECT * FROM {full_name:i}",
                    {
                        "full_name": bp.full_name,
                    },
                )
            except SnowDDLExecuteError as e:
                self.engine.logger.debug(
                    f"View [{bp.full_name}] caused describe error [{e.snow_exc.errno}]: {e.snow_exc.raw_msg}"
                )
            else:
                # Comments on views are broken and must be applied separately
                if bp.comment != row["comment"]:
                    self.engine.execute_safe_ddl(
                        "COMMENT ON VIEW {full_name:i} IS {comment}",
                        {
                            "full_name": bp.full_name,
                            "comment": bp.comment if bp.comment else "",
                        },
                    )

                    return ResolveResult.ALTER
                else:
                    return ResolveResult.NOCHANGE
        else:
            self.engine.logger.debug(f"View [{bp.full_name}] text did not match")

        # Replace view if we got here
        self.engine.execute_safe_ddl(query)

        # Comments on views are broken and must be applied separately
        if bp.comment:
            self.engine.execute_safe_ddl(
                "COMMENT ON VIEW {full_name:i} IS {comment}",
                {
                    "full_name": bp.full_name,
                    "comment": bp.comment,
                },
            )

        return ResolveResult.REPLACE

    def drop_object(self, row: dict):
        self.engine.execute_safe_ddl(
            "DROP VIEW {database:i}.{schema:i}.{view_name:i}",
            {
                "database": row["database"],
                "schema": row["schema"],
                "view_name": row["name"],
            },
        )

        return ResolveResult.DROP

    def _build_create_view(self, bp: ViewBlueprint):
        query = self.engine.query_builder()

        query.append("CREATE OR REPLACE")

        if bp.is_secure:
            query.append("SECURE")

        query.append(
            "VIEW {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        if bp.columns:
            query.append_nl("(")

            for idx, c in enumerate(bp.columns):
                query.append_nl(
                    "    {comma:r}{col_name:i}",
                    {
                        "comma": "  " if idx == 0 else ", ",
                        "col_name": c.name,
                    },
                )

                if c.comment:
                    query.append(
                        "COMMENT {col_comment}",
                        {
                            "col_comment": c.comment,
                        },
                    )

            query.append_nl(")")

        query.append_nl("COPY GRANTS")
        query.append_nl("AS")
        query.append_nl(bp.text)

        return query



from typing import Union

from .ident import AbstractIdent, DatabaseIdent, SchemaIdent
from .object_type import ObjectType
from ..model import BaseModelWithConfig


class Grant(BaseModelWithConfig):
    privilege: str
    on: ObjectType
    name: AbstractIdent

    def __eq__(self, other):
        if not isinstance(other, Grant):
            raise NotImplementedError

        if self.privilege != other.privilege:
            return False

        if self.on.singular_for_grant != other.on.singular_for_grant:
            return False

        if self.name != other.name:
            return False

        return True


class AccountGrant(BaseModelWithConfig):
    privilege: str


class FutureGrant(BaseModelWithConfig):
    privilege: str
    on_future: ObjectType
    in_parent: ObjectType
    name: Union[DatabaseIdent, SchemaIdent]



from typing import Union


def coalesce(val, default=None):
    return default if val is None else val


def compare_dynamic_param_value(bp_value: Union[bool, int, float, str], existing_value: str) -> bool:
    if isinstance(bp_value, bool):
        if bp_value is True and existing_value == "true":
            return True

        if bp_value is False and existing_value == "false":
            return True

    elif isinstance(bp_value, int):
        if bp_value == int(existing_value):
            return True

    elif isinstance(bp_value, float):
        if bp_value == float(existing_value):
            return True

    else:
        if str(bp_value) == existing_value:
            return True

    return False


def dtypes_from_arguments(arguments: str) -> str:
    arguments = arguments.replace("DEFAULT ", "")
    arguments = arguments.translate(str.maketrans("", "", "[] "))

    start_dtypes_idx = arguments.index("(")
    finish_dtypes_idx = arguments.index(")")

    return ",".join(a for a in arguments[start_dtypes_idx + 1 : finish_dtypes_idx].split(","))



from snowddl.blueprint import DynamicTableBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType


class DynamicTableResolver(AbstractSchemaObjectResolver):
    # Dynamic tables are available for all accounts during preview, including STANDARD edition
    # skip_min_edition = Edition.ENTERPRISE
    skip_on_empty_blueprints = True

    def get_object_type(self) -> ObjectType:
        return ObjectType.DYNAMIC_TABLE

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW DYNAMIC TABLES IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            existing_objects[f"{r['database_name']}.{r['schema_name']}.{r['name']}"] = {
                "database": r["database_name"],
                "schema": r["schema_name"],
                "name": r["name"],
                "owner": r["owner"],
                # Extract SQL query text only, skip the initial "CREATE DYNAMIC TABLE ..." part
                # Snowflake modifies original SQL text in this column, it cannot be compared directly
                "text": r["text"].partition("\nAS\n")[2],
                "target_lag": r["target_lag"],
                "refresh_mode": r["refresh_mode"],
                "warehouse": r["warehouse"],
                "comment": r["comment"] if r["comment"] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(DynamicTableBlueprint)

    def create_object(self, bp: DynamicTableBlueprint):
        query = self.engine.query_builder()
        query.append("CREATE OR REPLACE")

        if bp.is_transient:
            query.append("TRANSIENT")

        query.append(
            "DYNAMIC TABLE {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        query.append(self._build_common_dynamic_table_sql(bp))

        self.engine.execute_safe_ddl(query)

        return ResolveResult.CREATE

    def compare_object(self, bp: DynamicTableBlueprint, row: dict):
        result = ResolveResult.NOCHANGE

        if bp.text != row["text"]:
            query = self.engine.query_builder()
            query.append("CREATE OR REPLACE")

            if bp.is_transient:
                query.append("TRANSIENT")

            query.append(
                "DYNAMIC TABLE {full_name:i}",
                {
                    "full_name": bp.full_name,
                },
            )

            query.append(self._build_common_dynamic_table_sql(bp))
            self.engine.execute_unsafe_ddl(query)

            return ResolveResult.REPLACE

        if bp.target_lag != row["target_lag"]:
            self.engine.execute_safe_ddl(
                "ALTER DYNAMIC TABLE {full_name:i} SET TARGET_LAG = {target_lag}",
                {
                    "full_name": bp.full_name,
                    "target_lag": bp.target_lag,
                },
            )

            result = ResolveResult.ALTER

        if str(bp.warehouse) != row["warehouse"]:
            self.engine.execute_safe_ddl(
                "ALTER DYNAMIC TABLE {full_name:i} SET WAREHOUSE = {warehouse:i}",
                {
                    "full_name": bp.full_name,
                    "warehouse": bp.warehouse,
                },
            )

            result = ResolveResult.ALTER

        if bp.comment != row["comment"]:
            self.engine.execute_safe_ddl(
                "ALTER DYNAMIC TABLE {full_name:i} SET COMMENT = {comment}",
                {
                    "full_name": bp.full_name,
                    "comment": bp.comment,
                },
            )

            result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl(
            "DROP DYNAMIC TABLE {database:i}.{schema:i}.{name:i}",
            {
                "database": row["database"],
                "schema": row["schema"],
                "name": row["name"],
            },
        )

        return ResolveResult.DROP

    def _build_common_dynamic_table_sql(self, bp: DynamicTableBlueprint):
        query = self.engine.query_builder()

        if bp.columns:
            query.append_nl("(")

            for idx, c in enumerate(bp.columns):
                query.append_nl(
                    "    {comma:r}{col_name:i}",
                    {
                        "comma": "  " if idx == 0 else ", ",
                        "col_name": c.name,
                    },
                )

                if c.comment:
                    query.append(
                        "COMMENT {col_comment}",
                        {
                            "col_comment": c.comment,
                        },
                    )

            query.append_nl(")")

        query.append_nl(
            "TARGET_LAG = {target_lag}",
            {
                "target_lag": bp.target_lag,
            },
        )

        query.append_nl(
            "WAREHOUSE = {warehouse:i}",
            {
                "warehouse": bp.warehouse,
            },
        )

        if bp.refresh_mode:
            query.append_nl(
                "REFRESH_MODE = {refresh_mode}",
                {
                    "refresh_mode": bp.refresh_mode,
                },
            )

        if bp.initialize:
            query.append_nl(
                "INITIALIZE = {initialize}",
                {
                    "initialize": bp.initialize,
                },
            )

        if bp.cluster_by:
            query.append_nl(
                "CLUSTER BY ({cluster_by:r})",
                {
                    "cluster_by": bp.cluster_by,
                },
            )

        if bp.retention_time is not None:
            query.append_nl(
                "DATA_RETENTION_TIME_IN_DAYS = {retention_time:d}",
                {
                    "retention_time": bp.retention_time,
                },
            )

        if bp.comment:
            query.append_nl(
                "COMMENT = {comment}",
                {
                    "comment": bp.comment,
                },
            )

        query.append_nl("AS")
        query.append_nl(bp.text)

        return query



from re import compile, IGNORECASE

from snowddl.blueprint import (
    HybridTableBlueprint,
    IndexReference,
    TableColumn,
    ForeignKeyReference,
    DataType,
    Ident,
    SchemaObjectIdent,
    build_schema_object_ident,
)
from snowddl.parser.abc_parser import AbstractParser, ParsedFile

col_type_re = compile(r"^(?P<type>[a-z0-9_]+(\((\d+)(,(\d+))?\))?)" r"(?P<not_null> NOT NULL)?$", IGNORECASE)


# fmt: off
hybrid_table_json_schema = {
    "type": "object",
    "properties": {
        "columns": {
            "type": "object",
            "minItems": 1,
            "items": {
                "anyOf": [
                    {
                        "type": "string"
                    },
                    {
                        "type": "object",
                        "properties": {
                            "type": {
                                "type": "string"
                            },
                            "default": {
                                "type": "string"
                            },
                            "default_sequence": {
                                "type": "string"
                            },
                            "expression": {
                                "type": "string"
                            },
                            "collate": {
                                "type": "string"
                            },
                            "comment": {
                                "type": "string"
                            }
                        },
                        "required": ["type"],
                        "additionalProperties": False
                    }
                ]
            }
        },
        "comment": {
            "type": "string"
        },
        "primary_key": {
            "type": "array",
            "items": {
                "type": "string"
            },
            "minItems": 1
        },
        "unique_keys": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {
                    "type": "string"
                },
                "minItems": 1,
            }
        },
        "foreign_keys": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "columns": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        },
                        "minItems": 1
                    },
                    "ref_table": {
                        "type": "string"
                    },
                    "ref_columns": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        },
                        "minItems": 1
                    }
                },
                "additionalProperties": False,
                "required": ["columns", "ref_table", "ref_columns"],
            },
            "minItems": 1
        },
        "indexes": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "columns": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        },
                        "minItems": 1
                    },
                    "include": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        },
                        "minItems": 1
                    },
                },
                "additionalProperties": False,
                "required": ["columns"],
            },
            "minItems": 1
        },
    },
    "additionalProperties": False,
    "required": ["columns", "primary_key"],
}
# fmt: on


class HybridTableParser(AbstractParser):
    def load_blueprints(self):
        self.parse_schema_object_files("hybrid_table", hybrid_table_json_schema, self.process_table)

    def process_table(self, f: ParsedFile):
        column_blueprints = []

        for col_name, col in f.params["columns"].items():
            # Short syntax
            if isinstance(col, str):
                col = {"type": col}

            m = col_type_re.match(col["type"])

            if not m:
                raise ValueError(
                    f"Incorrect short syntax for column [{col_name}] in table [{f.database}.{f.schema}.{f.name}]: {col['type']}"
                )

            # Default for sequence
            if col.get("default_sequence"):
                col_default = build_schema_object_ident(self.env_prefix, col["default_sequence"], f.database, f.schema)
            else:
                col_default = col.get("default")

            column_blueprints.append(
                TableColumn(
                    name=Ident(col_name),
                    type=DataType(m.group("type")),
                    not_null=bool(m.group("not_null")),
                    default=col_default,
                    expression=col.get("expression"),
                    collate=col.get("collate").lower() if col.get("collate") else None,
                    comment=col.get("comment"),
                )
            )

        primary_key = [Ident(c) for c in f.params.get("primary_key")]
        unique_keys = None
        foreign_keys = None
        indexes = None
        depends_on = set()

        if f.params.get("unique_keys"):
            unique_keys = [[Ident(c) for c in columns] for columns in f.params.get("unique_keys")]

        if f.params.get("foreign_keys"):
            foreign_keys = [
                ForeignKeyReference(
                    columns=[Ident(c) for c in fk["columns"]],
                    ref_table_name=build_schema_object_ident(self.env_prefix, fk["ref_table"], f.database, f.schema),
                    ref_columns=[Ident(c) for c in fk["ref_columns"]],
                )
                for fk in f.params.get("foreign_keys")
            ]

            depends_on = set(fk.ref_table_name for fk in foreign_keys)

        if f.params.get("indexes"):
            indexes = [
                IndexReference(
                    columns=[Ident(c) for c in idx["columns"]],
                    include=[Ident(c) for c in idx["include"]] if idx.get("include") else None,
                )
                for idx in f.params.get("indexes")
            ]

        bp = HybridTableBlueprint(
            full_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
            columns=column_blueprints,
            primary_key=primary_key,
            unique_keys=unique_keys,
            foreign_keys=foreign_keys,
            indexes=indexes,
            depends_on=depends_on,
            comment=f.params.get("comment"),
        )

        self.config.add_blueprint(bp)



from abc import ABC
from pydantic import BaseModel, ConfigDict


class BaseModelWithConfig(BaseModel, ABC):
    model_config = ConfigDict(
        allow_inf_nan=False,
        arbitrary_types_allowed=True,
        extra="forbid",
        validate_assignment=True,
    )



from io import BytesIO
from hashlib import md5
from pathlib import Path

from snowddl.blueprint import StageBlueprint, StageFileBlueprint
from snowddl.error import SnowDDLExecuteError
from snowddl.resolver.abc_resolver import AbstractResolver, ResolveResult, ObjectType


class StageFileResolver(AbstractResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.STAGE_FILE

    def get_existing_objects(self):
        existing_objects = {}

        for stage_bp in self.config.get_blueprints_by_type(StageBlueprint).values():
            if not stage_bp.upload_stage_files:
                continue

            try:
                cur = self.engine.execute_meta(
                    "LIST @{stage_name:i}",
                    {
                        "stage_name": stage_bp.full_name,
                    },
                )
            except SnowDDLExecuteError as e:
                # Stage does not exist or not authorized
                # Skip this error during planning
                if e.snow_exc.errno == 2003:
                    continue
                else:
                    raise

            all_files = {}
            all_hashes = {}

            for r in cur:
                path = Path(r["name"])

                if path.suffix == ".md5":
                    all_hashes[path.with_suffix("").with_suffix("")] = path.suffixes[-2].lstrip(".")
                else:
                    all_files[path] = True

            for path in all_files:
                # Snowflake LIST commands adds stage name implicitly, which should be removed
                stage_path = f"/{path.relative_to(path.parts[0])}"
                full_name = f"{stage_bp.full_name}({stage_path})"

                # Snowflake LIST commands provides "md5" and "size", but it is not reliable due to encryption
                existing_objects[full_name] = {
                    "stage_name": stage_bp.full_name,
                    "stage_path": stage_path,
                    "original_md5": all_hashes.get(path, None),
                }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(StageFileBlueprint)

    def create_object(self, bp: StageFileBlueprint):
        self._upload_file(bp)
        self._upload_md5_marker(bp)

        return ResolveResult.CREATE

    def compare_object(self, bp: StageFileBlueprint, row: dict):
        if row["original_md5"] == self._md5_file(bp.local_path):
            return ResolveResult.NOCHANGE

        self._upload_file(bp)
        self._upload_md5_marker(bp)

        return ResolveResult.REPLACE

    def drop_object(self, row: dict):
        # One call deletes original file and MD5 marker in one go
        self.engine.execute_safe_ddl(
            "REMOVE @{stage_name:i}{stage_path:r}",
            {
                "stage_name": row["stage_name"],
                "stage_path": row["stage_path"],
            },
        )

        return ResolveResult.DROP

    def _upload_file(self, bp: StageFileBlueprint):
        self.engine.execute_safe_ddl(
            "PUT {local_path} @{stage_name:i}{stage_target:r} PARALLEL=1 OVERWRITE=TRUE AUTO_COMPRESS=FALSE",
            {
                "local_path": f"file://{bp.local_path}",
                "stage_name": bp.stage_name,
                "stage_target": Path(bp.stage_path).parent,
            },
        )

    def _upload_md5_marker(self, bp: StageFileBlueprint):
        # Placeholder path for PUT command, directory does not matter
        # Actual contents of marker pseudo-file is empty and come from zero-length BytesIO in file_stream
        md5_marker_path = Path(bp.local_path).name + f".{self._md5_file(bp.local_path)}.md5"

        self.engine.execute_safe_ddl(
            "PUT {local_path} @{stage_name:i}{stage_target:r} PARALLEL=1 OVERWRITE=TRUE AUTO_COMPRESS=FALSE",
            {
                "local_path": f"file://{md5_marker_path}",
                "stage_name": bp.stage_name,
                "stage_target": Path(bp.stage_path).parent,
            },
            file_stream=BytesIO(),
        )

    def _md5_file(self, local_path: str):
        hash_md5 = md5()

        with open(local_path, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)

        return hash_md5.hexdigest()

    def destroy(self):
        # No need to delete stage files explicitly, files are destroyed automatically when stage is gone
        pass



from snowddl.blueprint import ShareRoleBlueprint, Grant, build_grant_name_ident
from snowddl.resolver.abc_role_resolver import AbstractRoleResolver, ObjectType


class ShareRoleResolver(AbstractRoleResolver):
    def get_role_suffix(self):
        return self.config.SHARE_ROLE_SUFFIX

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(ShareRoleBlueprint)

    def get_existing_role_grants(self, role_name):
        # There is no way to detect grant of IMPORTED PRIVILEGES directly
        # SHOW GRANTS output is filled with whatever grants come with INBOUND SHARE
        # IMPORTED PRIVILEGES grant has to be emulated
        grants = []

        cur = self.engine.execute_meta(
            "SHOW GRANTS TO ROLE {role_name:i}",
            {
                "role_name": role_name,
            },
        )

        for r in cur:
            # Convert DATABASE.USAGE grant into IMPORTED PRIVILEGES
            # Ignore everything else
            if r["privilege"] == "USAGE" and r["granted_on"] == "DATABASE":
                grants.append(
                    Grant(
                        privilege="IMPORTED PRIVILEGES",
                        on=ObjectType.DATABASE,
                        name=build_grant_name_ident(ObjectType.DATABASE, r["name"]),
                    )
                )

                break

        return role_name, grants, [], []



from snowddl.blueprint import StreamBlueprint, SchemaObjectIdent, ObjectType, build_schema_object_ident
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


# fmt: off
stream_json_schema = {
    "type": "object",
    "properties": {
        "object_type": {
            "type": "string"
        },
        "object_name": {
            "type": "string"
        },
        "append_only": {
            "type": "boolean"
        },
        "insert_only": {
            "type": "boolean"
        },
        "show_initial_rows": {
            "type": "boolean"
        },
        "comment": {
            "type": "string"
        }
    },
    "required": ["object_type", "object_name"],
    "additionalProperties": False
}
# fmt: on


class StreamParser(AbstractParser):
    def load_blueprints(self):
        self.parse_schema_object_files("stream", stream_json_schema, self.process_stream)

    def process_stream(self, f: ParsedFile):
        bp = StreamBlueprint(
            full_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
            object_type=ObjectType[f.params["object_type"]],
            object_name=build_schema_object_ident(self.env_prefix, f.params["object_name"], f.database, f.schema),
            append_only=f.params.get("append_only"),
            insert_only=f.params.get("insert_only"),
            show_initial_rows=f.params.get("show_initial_rows"),
            comment=f.params.get("comment"),
        )

        self.config.add_blueprint(bp)



from snowddl.blueprint import SchemaBlueprint, SchemaIdent, Grant, ObjectType, build_role_ident
from snowddl.parser.abc_parser import AbstractParser
from snowddl.parser.database import database_json_schema


# fmt: off
schema_json_schema = {
    "type": "object",
    "properties": {
        "permission_model": {
            "type": "string",
        },
        "is_transient": {
            "type": "boolean"
        },
        "retention_time": {
            "type": "integer"
        },
        "is_sandbox": {
            "type": "boolean"
        },
        "owner_schema_read": {
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "owner_schema_write": {
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "owner_share_read": {
            "share_read": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
        },
        "owner_integration_usage": {
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "owner_warehouse_usage": {
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "owner_account_grants": {
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "owner_global_roles": {
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "comment": {
            "type": "string"
        }
    },
    "additionalProperties": False
}
# fmt: on


class SchemaParser(AbstractParser):
    def load_blueprints(self):
        for database_path in self.base_path.iterdir():
            if not database_path.is_dir():
                continue

            # Skip special sub-directories
            if database_path.name.startswith("__"):
                continue

            database_params = self.parse_single_file(database_path / "params.yaml", database_json_schema)

            for schema_path in database_path.iterdir():
                if not schema_path.is_dir():
                    continue

                schema_params = self.parse_single_file(schema_path / "params.yaml", schema_json_schema)

                combined_params = {
                    "is_transient": database_params.get("is_transient", False) or schema_params.get("is_transient", False),
                    "retention_time": schema_params.get("retention_time"),
                    "is_sandbox": database_params.get("is_sandbox", False) or schema_params.get("is_sandbox", False),
                }

                database_name = database_path.name.upper()
                schema_name = schema_path.name.upper()

                # fmt: off
                database_permission_model_name = database_params.get("permission_model", self.config.DEFAULT_PERMISSION_MODEL).upper()
                schema_permission_model_name = schema_params.get("permission_model", database_permission_model_name).upper()
                # fmt: on

                database_permission_model = self.config.get_permission_model(database_permission_model_name)
                schema_permission_model = self.config.get_permission_model(schema_permission_model_name)

                if database_permission_model.ruleset != schema_permission_model.ruleset:
                    raise ValueError(
                        f"Database [{database_name}] permission model ruleset does not match schema [{database_name}.{schema_name}] permission model ruleset"
                    )

                if not schema_permission_model.ruleset.create_schema_owner_role:
                    for k in schema_params:
                        if k.startswith("owner_"):
                            raise ValueError(
                                f"Cannot use parameter [{k}] for schema [{database_name}.{schema_name}], it should be configured on database level"
                            )

                owner_additional_grants = []
                owner_additional_account_grants = []

                for full_schema_name in schema_params.get("owner_schema_read", []):
                    owner_additional_grants.append(self.build_schema_role_grant(full_schema_name, self.config.READ_ROLE_TYPE))

                for full_schema_name in schema_params.get("owner_schema_write", []):
                    owner_additional_grants.append(self.build_schema_role_grant(full_schema_name, self.config.WRITE_ROLE_TYPE))

                for share_name in schema_params.get("owner_share_read", []):
                    owner_additional_grants.append(self.build_share_role_grant(share_name))
                    self.config.add_blueprint(self.build_share_role_blueprint(share_name))

                for integration_name in schema_params.get("owner_integration_usage", []):
                    owner_additional_grants.append(self.build_integration_usage_grant(integration_name))

                for warehouse_name in schema_params.get("owner_warehouse_usage", []):
                    owner_additional_grants.append(self.build_warehouse_role_grant(warehouse_name, self.config.USAGE_ROLE_TYPE))

                for account_grant in schema_params.get("owner_account_grants", []):
                    owner_additional_account_grants.append(self.build_account_grant(account_grant))

                for global_role_name in schema_params.get("owner_global_roles", []):
                    owner_additional_grants.append(self.build_global_role_grant(global_role_name))

                bp = SchemaBlueprint(
                    full_name=SchemaIdent(self.env_prefix, database_name, schema_name),
                    permission_model=schema_permission_model_name,
                    is_transient=combined_params.get("is_transient", False),
                    retention_time=combined_params.get("retention_time", None),
                    is_sandbox=combined_params.get("is_sandbox", False),
                    owner_additional_grants=owner_additional_grants,
                    owner_additional_account_grants=owner_additional_account_grants,
                    comment=schema_params.get("comment", None),
                )

                self.config.add_blueprint(bp)

    def build_schema_role_grant(self, full_schema_name, role_type):
        database, schema = full_schema_name.split(".")

        return Grant(
            privilege="USAGE",
            on=ObjectType.ROLE,
            name=build_role_ident(self.env_prefix, database, schema, role_type, self.config.SCHEMA_ROLE_SUFFIX),
        )



from snowddl.blueprint import ResourceMonitorBlueprint
from snowddl.resolver.abc_resolver import AbstractResolver, ResolveResult, ObjectType


class ResourceMonitorResolver(AbstractResolver):
    skip_on_empty_blueprints = True

    def get_object_type(self) -> ObjectType:
        return ObjectType.RESOURCE_MONITOR

    def get_existing_objects(self):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW RESOURCE MONITORS LIKE {env_prefix:ls}",
            {
                "env_prefix": self.config.env_prefix,
            },
        )

        for r in cur:
            if r["owner"] != self.engine.context.current_role:
                continue

            triggers = {}

            triggers.update(self._parse_triggers(r["notify_at"], "NOTIFY"))
            triggers.update(self._parse_triggers(r["suspend_at"], "SUSPEND"))
            triggers.update(self._parse_triggers(r["suspend_immediately_at"], "SUSPEND_IMMEDIATE"))

            existing_objects[r["name"]] = {
                "name": r["name"],
                "credit_quota": int(float(r["credit_quota"])),
                "frequency": r["frequency"],
                "notify_at": r["notify_at"],
                "suspend_at": r["suspend_at"],
                "suspend_immediately_at": r["suspend_immediately_at"],
                "triggers": triggers,
                "comment": r["comment"] if r["comment"] else None,
            }

        return existing_objects

    def _parse_triggers(self, val, action):
        triggers = {}

        if val is None:
            return triggers

        for threshold in val.split(","):
            threshold = threshold.strip(" %")
            triggers[int(threshold)] = action

        return triggers

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(ResourceMonitorBlueprint)

    def create_object(self, bp: ResourceMonitorBlueprint):
        query = self.engine.query_builder()

        query.append(
            "CREATE RESOURCE MONITOR {name:i}",
            {
                "name": bp.full_name,
            },
        )

        query.append_nl(
            "CREDIT_QUOTA = {credit_quota:d}",
            {
                "credit_quota": bp.credit_quota,
            },
        )

        query.append_nl(
            "FREQUENCY = {frequency}",
            {
                "frequency": bp.frequency,
            },
        )

        query.append_nl("START_TIMESTAMP = IMMEDIATELY")
        query.append_nl("TRIGGERS")

        for threshold, action in bp.triggers.items():
            query.append_nl(
                "ON {threshold:d} PERCENT DO {action:r}",
                {
                    "threshold": threshold,
                    "action": action,
                },
            )

        self.engine.execute_unsafe_ddl(query, condition=self.engine.settings.execute_resource_monitor)

        return ResolveResult.CREATE

    def compare_object(self, bp: ResourceMonitorBlueprint, row: dict):
        result = ResolveResult.NOCHANGE

        if bp.credit_quota != row["credit_quota"]:
            self.engine.execute_unsafe_ddl(
                "ALTER RESOURCE MONITOR {name:i} SET CREDIT_QUOTA = {credit_quota:d}",
                {
                    "name": bp.full_name,
                    "credit_quota": bp.credit_quota,
                },
                condition=self.engine.settings.execute_resource_monitor,
            )

            result = ResolveResult.ALTER

        if bp.frequency != row["frequency"]:
            self.engine.execute_unsafe_ddl(
                "ALTER RESOURCE MONITOR {name:i} SET FREQUENCY = {frequency} START_TIMESTAMP = IMMEDIATELY",
                {
                    "name": bp.full_name,
                    "frequency": bp.frequency,
                },
                condition=self.engine.settings.execute_resource_monitor,
            )

            result = ResolveResult.ALTER

        if bp.triggers != row["triggers"]:
            query = self.engine.query_builder()

            query.append(
                "ALTER RESOURCE MONITOR {name:i}",
                {
                    "name": bp.full_name,
                },
            )

            query.append_nl("TRIGGERS")

            for threshold, action in bp.triggers.items():
                query.append_nl(
                    "ON {threshold:d} PERCENT DO {action:r}",
                    {
                        "threshold": threshold,
                        "action": action,
                    },
                )

            self.engine.execute_unsafe_ddl(query, condition=self.engine.settings.execute_resource_monitor)

            result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl(
            "DROP RESOURCE MONITOR {name:i}",
            {
                "name": row["name"],
            },
            condition=self.engine.settings.execute_resource_monitor,
        )

        return ResolveResult.DROP



from snowddl.blueprint import (
    ObjectType,
    PermissionModel,
    PermissionModelCreateGrant,
    PermissionModelFutureGrant,
    PermissionModelRuleset,
)
from snowddl.parser.abc_parser import AbstractParser


# fmt: off
permission_model_json_schema = {
    "type": "object",
    "additionalProperties": {
        "type": "object",
        "properties": {
            "inherit_from": {
                "type": "string",
            },
            "ruleset": {
                "type": "string",
            },
            "owner_create_grants": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
            "owner_future_grants": {
                "type": "object",
                "additionalProperties": {
                    "type": "array",
                    "items": {
                        "type": "string"
                    },
                    "minItems": 1
                }
            },
            "write_future_grants": {
                "type": "object",
                "additionalProperties": {
                    "type": "array",
                    "items": {
                        "type": "string"
                    },
                    "minItems": 1
                }
            },
            "read_future_grants": {
                "type": "object",
                "additionalProperties": {
                    "type": "array",
                    "items": {
                        "type": "string"
                    },
                    "minItems": 1
                }
            },
        },
        "additionalProperties": False,
    },
}


class PermissionModelParser(AbstractParser):
    def load_blueprints(self):
        # This is a special parser that does not load any blueprints, but it loads permission models instead
        pass

    def load_permission_models(self):
        custom_permission_models = self.parse_single_file(self.base_path / "permission_model.yaml", permission_model_json_schema)

        for name, params in custom_permission_models.items():
            self.build_permission_model(name, params)

    def build_permission_model(self, name, params):
        # Possibly inherit model from another model to reduce code repetition
        if params.get("inherit_from"):
            model = self.config.get_permission_model(params.get("inherit_from").upper()).model_copy(deep=True)
        else:
            model = PermissionModel()

        # Ruleset
        if params.get("ruleset"):
            model.ruleset = PermissionModelRuleset[params.get("ruleset").upper()]

        # Owner create grants
        for object_type in params.get("owner_create_grants", []):
            grant = PermissionModelCreateGrant(on=ObjectType[str(object_type).upper()])

            if grant not in model.owner_future_grants:
                model.owner_create_grants.append(grant)

        # Owner future grants
        for object_type, privileges in params.get("owner_future_grants", {}).items():
            for p in privileges:
                grant = PermissionModelFutureGrant(privilege=p.upper(), on=ObjectType[str(object_type).upper()])

                if grant not in model.owner_future_grants:
                    model.owner_future_grants.append(grant)

        # Write future grants
        for object_type, privileges in params.get("write_future_grants", {}).items():
            for p in privileges:
                if p.upper() == "OWNERSHIP":
                    raise ValueError(
                        f"Cannot assign OWNERSHIP future grant to WRITE role while defining permission model [{name}]"
                    )

                grant = PermissionModelFutureGrant(privilege=p.upper(), on=ObjectType[str(object_type).upper()])

                if grant not in model.write_future_grants:
                    model.write_future_grants.append(grant)

        # Read future grants
        for object_type, privileges in params.get("read_future_grants", {}).items():
            for p in privileges:
                if p.upper() == "OWNERSHIP":
                    raise ValueError(
                        f"Cannot assign OWNERSHIP future grant to READ role while defining permission model [{name}]"
                    )

                grant = PermissionModelFutureGrant(privilege=p.upper(), on=ObjectType[str(object_type).upper()])

                if grant not in model.read_future_grants:
                    model.read_future_grants.append(grant)

        self.config.add_permission_model(name.upper(), model)



from snowddl.blueprint import PipeBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType
from snowddl.resolver._utils import coalesce


class PipeResolver(AbstractSchemaObjectResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.PIPE

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW PIPES IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            existing_objects[f"{r['database_name']}.{r['schema_name']}.{r['name']}"] = {
                "database": r["database_name"],
                "schema": r["schema_name"],
                "name": r["name"],
                "definition": r["definition"],
                "integration": r["integration"],
                "pattern": r["pattern"],
                "comment": r["comment"],
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(PipeBlueprint)

    def create_object(self, bp: PipeBlueprint):
        common_query = self._build_common_pipe_sql(bp)
        create_query = self.engine.query_builder()

        create_query.append(
            "CREATE PIPE {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        create_query.append(common_query)

        self.engine.execute_safe_ddl(create_query)

        self.engine.execute_safe_ddl(
            "COMMENT ON PIPE {full_name:i} IS {comment}",
            {
                "full_name": bp.full_name,
                "comment": common_query.add_short_hash(bp.comment),
            },
        )

        return ResolveResult.CREATE

    def compare_object(self, bp: PipeBlueprint, row: dict):
        common_query = self._build_common_pipe_sql(bp)

        if not common_query.compare_short_hash(row["comment"]):
            replace_query = self.engine.query_builder()

            replace_query.append(
                "CREATE OR REPLACE PIPE {full_name:i}",
                {
                    "full_name": bp.full_name,
                },
            )

            replace_query.append(common_query)

            self.engine.execute_unsafe_ddl(replace_query)

            self.engine.execute_unsafe_ddl(
                "COMMENT ON PIPE {full_name:i} IS {comment}",
                {
                    "full_name": bp.full_name,
                    "comment": common_query.add_short_hash(bp.comment),
                },
            )

            return ResolveResult.REPLACE

        return ResolveResult.NOCHANGE

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl(
            "DROP PIPE {database:i}.{schema:i}.{name:i}",
            {
                "database": row["database"],
                "schema": row["schema"],
                "name": row["name"],
            },
        )

        return ResolveResult.DROP

    def _build_common_pipe_sql(self, bp: PipeBlueprint):
        query = self.engine.query_builder()

        query.append_nl(
            "AUTO_INGEST = {auto_ingest:b}",
            {
                "auto_ingest": bp.auto_ingest,
            },
        )

        if bp.aws_sns_topic:
            query.append_nl(
                "AWS_SNS_TOPIC = {aws_sns_topic}",
                {
                    "aws_sns_topic": bp.aws_sns_topic,
                },
            )

        if bp.integration:
            query.append_nl(
                "INTEGRATION = {integration:i}",
                {
                    "integration": bp.integration,
                },
            )

        if bp.error_integration:
            query.append_nl(
                "ERROR_INTEGRATION = {error_integration:i}",
                {
                    "error_integration": bp.error_integration,
                },
            )

        if bp.comment:
            query.append_nl(
                "COMMENT = {comment}",
                {
                    "comment": bp.comment,
                },
            )

        query.append_nl("AS")
        query.append_nl(self._build_copy_into(bp))

        return query

    def _build_copy_into(self, bp: PipeBlueprint):
        query = self.engine.query_builder()

        query.append(
            "COPY INTO {table_name:i}",
            {
                "table_name": bp.copy_table_name,
            },
        )

        if bp.copy_transform:
            query.append(" (")

            for idx, col_name in enumerate(bp.copy_transform):
                query.append_nl(
                    "    {comma:r}{col_name:i}",
                    {
                        "comma": "  " if idx == 0 else ", ",
                        "col_name": col_name,
                    },
                )

            query.append_nl(") FROM (")
            query.append_nl("    SELECT")

            for idx, col_expr in enumerate(bp.copy_transform.values()):
                query.append_nl(
                    "        {comma:r}{col_expr:r}",
                    {
                        "comma": "  " if idx == 0 else ", ",
                        "col_expr": col_expr,
                    },
                )

            query.append_nl(
                "    FROM @{stage_name:i}{path:r}",
                {
                    "stage_name": bp.copy_stage_name,
                    "path": coalesce(bp.copy_path, ""),
                },
            )

            query.append_nl(")")
        else:
            query.append_nl(
                "FROM @{stage_name:i}{path:r}",
                {
                    "stage_name": bp.copy_stage_name,
                    "path": coalesce(bp.copy_path, ""),
                },
            )

        if bp.copy_pattern:
            query.append_nl(
                "PATTERN = {pattern}",
                {
                    "pattern": bp.copy_pattern,
                },
            )

        if bp.copy_file_format:
            query.append_nl(
                "FILE_FORMAT = {file_format:i}",
                {
                    "file_format": bp.copy_file_format,
                },
            )

        if bp.copy_match_by_column_name:
            query.append_nl(
                "MATCH_BY_COLUMN_NAME = {match_by_column_name}",
                {
                    "match_by_column_name": bp.copy_match_by_column_name.upper(),
                },
            )

        if bp.copy_include_metadata:
            query.append_nl("INCLUDE_METADATA = (")

            for idx, (table_col_name, metadata_col_name) in enumerate(bp.copy_include_metadata.items()):
                query.append_nl(
                    "    {comma:r}{table_col_name:i} = {metadata_col_name:i}",
                    {
                        "comma": "  " if idx == 0 else ", ",
                        "table_col_name": table_col_name,
                        "metadata_col_name": metadata_col_name,
                    },
                )

            query.append_nl(")")

        if bp.copy_options:
            for k, v in bp.copy_options.items():
                query.append_nl(
                    "{option_name:r} = {option_value:dp}",
                    {
                        "option_name": k,
                        "option_value": v,
                    },
                )

        return query



from abc import ABC, abstractmethod

from snowddl.converter.abc_converter import AbstractConverter


class AbstractSchemaObjectConverter(AbstractConverter, ABC):
    def get_existing_objects(self):
        existing_objects = {}

        # Process schemas in parallel
        for schema_objects in self.engine.executor.map(
            self.get_existing_objects_in_schema, self.engine.schema_cache.schemas.values()
        ):
            existing_objects.update(schema_objects)

        return existing_objects

    @abstractmethod
    def get_existing_objects_in_schema(self, schema: dict):
        pass



from snowddl.blueprint import TableBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractResolver, ResolveResult, ObjectType


class CloneTableResolver(AbstractResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.CLONE_TABLE

    def get_blueprints(self):
        return {}

    def get_existing_objects(self):
        databases_for_clone = self.get_databases_for_clone()
        schemas_for_clone = {}
        tables_for_clone = {}

        # Get schemas for clone in parallel
        for schema_objects in self.engine.executor.map(self.get_schemas_for_clone, databases_for_clone.values()):
            schemas_for_clone.update(schema_objects)

        # Get tables for clone in parallel
        for table_objects in self.engine.executor.map(self.get_tables_for_clone, schemas_for_clone.values()):
            tables_for_clone.update(table_objects)

        return tables_for_clone

    def get_databases_for_clone(self):
        databases_for_clone = {}

        cur = self.engine.execute_meta("SHOW DATABASES")

        for r in cur:
            # Skip shares
            if r["origin"]:
                continue

            # Skip databases without destination for cloning
            if f"{self.config.env_prefix}{r['name']}" not in self.engine.schema_cache.databases:
                continue

            databases_for_clone[r["name"]] = {
                "database": r["name"],
            }

        return databases_for_clone

    def get_schemas_for_clone(self, database):
        schemas_for_clone = {}

        cur = self.engine.execute_meta(
            "SHOW SCHEMAS IN DATABASE {database:i}",
            {
                "database": database["database"],
            },
        )

        for r in cur:
            # Skip INFORMATION_SCHEMA
            if r["name"] == "INFORMATION_SCHEMA":
                continue

            # Skip schemas without destination for cloning
            if f"{self.config.env_prefix}{r['database_name']}.{r['name']}" not in self.engine.schema_cache.schemas:
                continue

            schemas_for_clone[f"{r['database_name']}.{r['name']}"] = {
                "database": r["database_name"],
                "schema": r["name"],
            }

        return schemas_for_clone

    def get_tables_for_clone(self, schema):
        tables_for_clone = {}

        cur = self.engine.execute_meta(
            "SHOW TABLES IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            # Skip other table types
            if (
                r.get("is_external") == "Y"
                or r.get("is_event") == "Y"
                or r.get("is_hybrid") == "Y"
                or r.get("is_iceberg") == "Y"
                or r.get("is_dynamic") == "Y"
            ):
                continue

            tables_for_clone[f"{self.config.env_prefix}{r['database_name']}.{r['schema_name']}.{r['name']}"] = {
                "database": r["database_name"],
                "schema": r["schema_name"],
                "name": r["name"],
                "is_transient": r["kind"] == "TRANSIENT",
            }

        return tables_for_clone

    def create_object(self, bp: TableBlueprint):
        return ResolveResult.NOCHANGE

    def compare_object(self, bp: TableBlueprint, row: dict):
        return ResolveResult.NOCHANGE

    def drop_object(self, row: dict):
        # Drop object is utilized since we have no blueprints for clones, but only existing objects
        # All such objects should be "dropped" from AbstractResolver perspective
        # TODO: find more elegant solution if cloning becomes popular

        query = self.engine.query_builder()
        query.append("CREATE")

        if row["is_transient"]:
            query.append("TRANSIENT")

        query.append(
            "TABLE IF NOT EXISTS {database_with_prefix:i}.{schema:i}.{table_name:i}",
            {
                "database_with_prefix": f"{self.config.env_prefix}{row['database']}",
                "schema": row["schema"],
                "table_name": row["name"],
            },
        )

        query.append_nl(
            "CLONE {database:i}.{schema:i}.{table_name:i}",
            {
                "database": row["database"],
                "schema": row["schema"],
                "table_name": row["name"],
            },
        )

        cur = self.engine.execute_clone(query)
        r = cur.fetchone()

        if str(r["status"]).endswith("successfully created."):
            return ResolveResult.CREATE

        return ResolveResult.NOCHANGE

    def destroy(self):
        pass

    def _is_skipped(self):
        if not self.engine.settings.clone_table:
            return True

        if super()._is_skipped():
            return True

        return False



from collections import defaultdict
from typing import Dict, Set, TYPE_CHECKING

from snowddl.blueprint import (
    ForeignKeyBlueprint,
    ObjectType,
    PrimaryKeyBlueprint,
    SchemaBlueprint,
    SchemaObjectBlueprint,
    StageFileBlueprint,
    UniqueKeyBlueprint,
)

if TYPE_CHECKING:
    from snowddl.engine import SnowDDLEngine


class IntentionCache:
    def __init__(self, engine: "SnowDDLEngine"):
        self.engine = engine

        self.drop_intention: Dict[ObjectType, Set[str]] = defaultdict(set)
        self.replace_intention: Dict[ObjectType, Set[str]] = defaultdict(set)

        self.invalid_name_warning: Dict[ObjectType, Set[str]] = defaultdict(set)

    def add_drop_intention(self, object_type: ObjectType, object_full_name: str):
        self.drop_intention[object_type].add(object_full_name)

    def add_replace_intention(self, object_type: ObjectType, object_full_name: str):
        self.replace_intention[object_type].add(object_full_name)

    def add_invalid_name_warning(self, object_type: ObjectType, object_full_name: str):
        self.invalid_name_warning[object_type].add(object_full_name)

    def check_drop_intention(self, object_type: ObjectType, object_full_name: str):
        return object_full_name in self.drop_intention[object_type]

    def check_parent_drop_intention(self, object_type: ObjectType, object_full_name: str):
        blueprint_cls = object_type.blueprint_cls
        object_full_name_parts = object_full_name.partition("(")[0].split(".")

        database_name = ".".join(object_full_name_parts[:1])
        schema_name = ".".join(object_full_name_parts[:2])
        schema_object_name = ".".join(object_full_name_parts[:3])

        # All schemas and schema objects are implicitly dropped by DATABASE
        if issubclass(blueprint_cls, (SchemaBlueprint, SchemaObjectBlueprint)) and (
            database_name in self.drop_intention[ObjectType.DATABASE]
        ):
            return True

        # All schema objects are implicitly dropped by SCHEMA
        if issubclass(blueprint_cls, SchemaObjectBlueprint) and (schema_name in self.drop_intention[ObjectType.SCHEMA]):
            return True

        # All stage files are implicitly dropped by STAGE
        if issubclass(blueprint_cls, StageFileBlueprint) and (schema_object_name in self.drop_intention[ObjectType.STAGE]):
            return True

        # All table constraints are implicitly dropped by various TABLE types which support constraints
        if issubclass(blueprint_cls, (ForeignKeyBlueprint, PrimaryKeyBlueprint, UniqueKeyBlueprint)) and (
            schema_object_name in self.drop_intention[ObjectType.TABLE]
            or schema_object_name in self.drop_intention[ObjectType.EXTERNAL_TABLE]
            or schema_object_name in self.drop_intention[ObjectType.HYBRID_TABLE]
        ):
            return True

        return False



from abc import ABC
from typing import Optional, List, Dict, Set, Union, TypeVar

from .column import (
    DynamicTableColumn,
    ExternalTableColumn,
    TableColumn,
    ViewColumn,
    ArgumentWithType,
    NameWithType,
    SearchOptimizationItem,
)
from .data_type import DataType
from .grant import AccountGrant, Grant, FutureGrant
from .ident import (
    AbstractIdent,
    Ident,
    AccountObjectIdent,
    DatabaseIdent,
    AccountIdent,
    OutboundShareIdent,
    SchemaIdent,
    SchemaObjectIdent,
    SchemaObjectIdentWithArgs,
    StageFileIdent,
    TableConstraintIdent,
)
from .object_type import ObjectType
from .reference import (
    AggregationPolicyReference,
    ForeignKeyReference,
    IndexReference,
    MaskingPolicyReference,
    ProjectionPolicyReference,
    RowAccessPolicyReference,
    TagReference,
)
from .stage import StageWithPath
from ..model import BaseModelWithConfig


class AbstractBlueprint(BaseModelWithConfig, ABC):
    full_name: AbstractIdent
    comment: Optional[str] = None


class SchemaObjectBlueprint(AbstractBlueprint, ABC):
    full_name: SchemaObjectIdent


class RoleBlueprint(AbstractBlueprint):
    full_name: AccountObjectIdent
    grants: List[Grant] = []
    account_grants: List[AccountGrant] = []
    future_grants: List[FutureGrant] = []


class DependsOnMixin(BaseModelWithConfig, ABC):
    depends_on: Set[AbstractIdent] = set()


class AccountParameterBlueprint(AbstractBlueprint):
    full_name: Ident
    value: Union[bool, float, int, str]


class AggregationPolicyBlueprint(SchemaObjectBlueprint):
    body: str
    references: List[AggregationPolicyReference]


class AlertBlueprint(SchemaObjectBlueprint):
    full_name: SchemaObjectIdent
    warehouse: AccountObjectIdent
    schedule: str
    condition: str
    action: str


class BusinessRoleBlueprint(RoleBlueprint):
    pass


class DatabaseBlueprint(AbstractBlueprint):
    full_name: DatabaseIdent
    permission_model: Optional[str] = None
    is_transient: Optional[bool] = None
    retention_time: Optional[int] = None
    is_sandbox: Optional[bool] = None
    owner_additional_grants: List[Grant] = []
    owner_additional_account_grants: List[AccountGrant] = []


class DatabaseRoleBlueprint(RoleBlueprint, DependsOnMixin):
    pass


class DynamicTableBlueprint(SchemaObjectBlueprint, DependsOnMixin):
    text: str
    columns: Optional[List[DynamicTableColumn]] = None
    target_lag: str
    warehouse: AccountObjectIdent
    refresh_mode: Optional[str] = None
    initialize: Optional[str] = None
    cluster_by: Optional[List[str]] = None
    is_transient: bool = False
    retention_time: Optional[int] = None


class EventTableBlueprint(SchemaObjectBlueprint):
    change_tracking: bool


class ExternalAccessIntegrationBlueprint(AbstractBlueprint):
    full_name: AccountObjectIdent
    allowed_network_rules: List[SchemaObjectIdent]
    allowed_api_authentication_integrations: Optional[List[Ident]] = None
    allowed_authentication_secrets: Optional[List[SchemaObjectIdent]] = None
    enabled: bool = True


class ExternalFunctionBlueprint(SchemaObjectBlueprint):
    full_name: SchemaObjectIdent
    arguments: List[NameWithType]
    returns: DataType
    api_integration: Ident
    url: str
    is_secure: bool = False
    is_strict: bool = False
    is_immutable: bool = False
    headers: Optional[Dict[str, str]] = None
    context_headers: Optional[List[Ident]] = None
    max_batch_rows: Optional[int] = None
    compression: Optional[str] = None
    request_translator: Optional[SchemaObjectIdent] = None
    response_translator: Optional[SchemaObjectIdent] = None


class ExternalTableBlueprint(SchemaObjectBlueprint):
    columns: Optional[List[ExternalTableColumn]] = None
    partition_by: Optional[List[Ident]] = None
    partition_type: Optional[str] = None
    location_stage: SchemaObjectIdent
    location_path: Optional[str] = None
    location_pattern: Optional[str] = None
    file_format: SchemaObjectIdent
    refresh_on_create: Optional[bool] = None
    auto_refresh: Optional[bool] = None
    aws_sns_topic: Optional[str] = None
    table_format: Optional[str] = None
    integration: Optional[Ident] = None


class FileFormatBlueprint(SchemaObjectBlueprint):
    type: str
    format_options: Optional[Dict[str, Union[bool, float, int, str, list]]] = None


class ForeignKeyBlueprint(SchemaObjectBlueprint):
    full_name: TableConstraintIdent
    table_name: SchemaObjectIdent
    columns: List[Ident]
    ref_table_name: SchemaObjectIdent
    ref_columns: List[Ident]


class FunctionBlueprint(SchemaObjectBlueprint):
    full_name: SchemaObjectIdentWithArgs
    language: str
    body: Optional[str] = None
    arguments: List[ArgumentWithType]
    returns: Union[DataType, List[NameWithType]]
    is_secure: bool = False
    is_strict: bool = False
    is_immutable: bool = False
    is_memoizable: bool = False
    runtime_version: Optional[str] = None
    imports: Optional[List[StageWithPath]] = None
    packages: Optional[List[str]] = None
    handler: Optional[str] = None
    external_access_integrations: Optional[List[AccountObjectIdent]] = None
    secrets: Optional[Dict[str, SchemaObjectIdent]] = None


class HybridTableBlueprint(SchemaObjectBlueprint, DependsOnMixin):
    columns: List[TableColumn]
    primary_key: List[Ident]
    unique_keys: Optional[List[List[Ident]]] = None
    foreign_keys: Optional[List[ForeignKeyReference]] = None
    indexes: Optional[List[IndexReference]] = None


class MaterializedViewBlueprint(SchemaObjectBlueprint):
    text: str
    columns: Optional[List[ViewColumn]] = None
    is_secure: bool = False
    cluster_by: Optional[List[str]] = None


class MaskingPolicyBlueprint(SchemaObjectBlueprint):
    arguments: List[NameWithType]
    returns: DataType
    body: str
    exempt_other_policies: bool = False
    references: List[MaskingPolicyReference]


class NetworkPolicyBlueprint(AbstractBlueprint):
    full_name: AccountObjectIdent
    allowed_ip_list: List[str] = []
    blocked_ip_list: List[str] = []


class NetworkRuleBlueprint(SchemaObjectBlueprint):
    type: str
    value_list: List[str]
    mode: str


class OutboundShareBlueprint(AbstractBlueprint):
    full_name: OutboundShareIdent
    accounts: List[AccountIdent] = []
    share_restrictions: bool = False
    grants: List[Grant] = []


class PipeBlueprint(SchemaObjectBlueprint):
    auto_ingest: bool
    copy_table_name: SchemaObjectIdent
    copy_stage_name: SchemaObjectIdent
    copy_path: Optional[str] = None
    copy_pattern: Optional[str] = None
    copy_transform: Optional[Dict[str, str]] = None
    copy_file_format: Optional[SchemaObjectIdent] = None
    copy_match_by_column_name: Optional[str] = None
    copy_include_metadata: Optional[Dict[Ident, Ident]] = None
    copy_options: Optional[Dict[str, Union[bool, float, int, str, list]]] = None
    aws_sns_topic: Optional[str] = None
    integration: Optional[Ident] = None
    error_integration: Optional[Ident] = None


class PrimaryKeyBlueprint(SchemaObjectBlueprint):
    table_name: SchemaObjectIdent
    columns: List[Ident]


class ProcedureBlueprint(SchemaObjectBlueprint):
    full_name: SchemaObjectIdentWithArgs
    language: str
    body: str
    arguments: List[ArgumentWithType]
    returns: Union[DataType, List[NameWithType]]
    is_strict: bool = False
    is_immutable: bool = False
    is_execute_as_caller: bool = False
    runtime_version: Optional[str] = None
    imports: Optional[List[StageWithPath]] = None
    packages: Optional[List[str]] = None
    handler: Optional[str] = None
    external_access_integrations: Optional[List[AccountObjectIdent]] = None
    secrets: Optional[Dict[str, SchemaObjectIdent]] = None


class ProjectionPolicyBlueprint(SchemaObjectBlueprint):
    body: str
    references: List[ProjectionPolicyReference]


class ResourceMonitorBlueprint(AbstractBlueprint):
    full_name: AccountObjectIdent
    credit_quota: int
    frequency: str
    triggers: Dict[int, str]


class RowAccessPolicyBlueprint(SchemaObjectBlueprint):
    arguments: List[NameWithType]
    body: str
    references: List[RowAccessPolicyReference]


class SchemaBlueprint(AbstractBlueprint):
    full_name: SchemaIdent
    permission_model: Optional[str] = None
    is_transient: Optional[bool] = None
    retention_time: Optional[int] = None
    is_sandbox: Optional[bool] = None
    owner_additional_grants: List[Grant] = []
    owner_additional_account_grants: List[AccountGrant] = []


class SchemaRoleBlueprint(RoleBlueprint, DependsOnMixin):
    pass


class SecretBlueprint(SchemaObjectBlueprint):
    type: str
    api_authentication: Optional[Ident] = None
    oauth_scopes: Optional[List[str]] = None
    oauth_refresh_token: Optional[str] = None
    oauth_refresh_token_expiry_time: Optional[str] = None
    username: Optional[str] = None
    password: Optional[str] = None
    secret_string: Optional[str] = None


class SequenceBlueprint(SchemaObjectBlueprint):
    start: int
    interval: int
    is_ordered: Optional[bool] = None


class ShareRoleBlueprint(RoleBlueprint):
    pass


class StageBlueprint(SchemaObjectBlueprint):
    url: Optional[str] = None
    storage_integration: Optional[Ident] = None
    encryption: Optional[Dict[str, Union[bool, float, int, str, list]]] = None
    directory: Optional[Dict[str, Union[bool, float, int, str, list]]] = None
    file_format: Optional[SchemaObjectIdent] = None
    copy_options: Optional[Dict[str, Union[bool, float, int, str, list]]] = None
    upload_stage_files: bool = False


class StageFileBlueprint(SchemaObjectBlueprint):
    full_name: StageFileIdent
    local_path: str
    stage_name: SchemaObjectIdent
    stage_path: str


class StreamBlueprint(SchemaObjectBlueprint):
    object_type: ObjectType
    object_name: SchemaObjectIdent
    append_only: Optional[bool] = None
    insert_only: Optional[bool] = None
    show_initial_rows: Optional[bool] = None


class TableBlueprint(SchemaObjectBlueprint):
    columns: List[TableColumn]
    cluster_by: Optional[List[str]] = None
    is_transient: bool = False
    retention_time: Optional[int] = None
    change_tracking: bool = False
    search_optimization: Union[bool, List[SearchOptimizationItem]] = False


class TagBlueprint(SchemaObjectBlueprint):
    full_name: SchemaObjectIdent
    references: List[TagReference]


class TaskBlueprint(SchemaObjectBlueprint, DependsOnMixin):
    body: str
    schedule: Optional[str] = None
    after: Optional[List[SchemaObjectIdent]] = None
    finalize: Optional[SchemaObjectIdent] = None
    when: Optional[str] = None
    warehouse: Optional[AccountObjectIdent] = None
    user_task_managed_initial_warehouse_size: Optional[str] = None
    config: Optional[str] = None
    allow_overlapping_execution: Optional[bool] = None
    session_params: Optional[Dict[str, Union[bool, float, int, str]]] = None
    user_task_timeout_ms: Optional[int] = None
    suspend_task_after_num_failures: Optional[int] = None
    error_integration: Optional[Ident] = None
    task_auto_retry_attempts: Optional[int] = None
    user_task_minimum_trigger_interval_in_seconds: Optional[int] = None


class TechnicalRoleBlueprint(RoleBlueprint):
    pass


class UniqueKeyBlueprint(SchemaObjectBlueprint):
    full_name: TableConstraintIdent
    table_name: SchemaObjectIdent
    columns: List[Ident]


class UserBlueprint(AbstractBlueprint):
    full_name: AccountObjectIdent
    login_name: str
    display_name: str
    first_name: Optional[str] = None
    last_name: Optional[str] = None
    email: Optional[str] = None
    disabled: bool = False
    business_roles: List[AccountObjectIdent] = []
    password: Optional[str] = None
    rsa_public_key: Optional[str] = None
    rsa_public_key_2: Optional[str] = None
    default_warehouse: Optional[AccountObjectIdent] = None
    default_namespace: Optional[Union[DatabaseIdent, SchemaIdent]] = None
    session_params: Optional[Dict[str, Union[bool, float, int, str]]] = None


class ViewBlueprint(SchemaObjectBlueprint, DependsOnMixin):
    text: str
    columns: Optional[List[ViewColumn]] = None
    is_secure: bool = False


class WarehouseBlueprint(AbstractBlueprint):
    full_name: AccountObjectIdent
    type: str
    size: str
    auto_suspend: int
    min_cluster_count: Optional[int] = None
    max_cluster_count: Optional[int] = None
    scaling_policy: Optional[str] = None
    resource_monitor: Optional[Union[Ident, AccountObjectIdent]] = None
    enable_query_acceleration: bool = False
    query_acceleration_max_scale_factor: Optional[int] = None
    warehouse_params: Optional[Dict[str, Union[bool, float, int, str]]] = None


T_Blueprint = TypeVar("T_Blueprint", bound=AbstractBlueprint)
